import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:flutter/foundation.dart';
import 'package:flutterx_http/flutterx_http.dart';
import 'package:flutterx_http/src/transform_request.dart';
import 'package:flutterx_utils/flutterx_utils.dart';
import 'package:http/http.dart';
import 'package:http/src/utils.dart'; // ignore: implementation_imports
import 'package:http_parser/http_parser.dart';

part 'x_http_exception.dart';
part 'x_http_extensions.dart';

/// Called before [request] is performed. You can apply modifications to source request before returning it since it
/// has not been finalized yet or even return a completely new request. The returned request will be used
typedef OnRequest = FutureOr<BaseRequest> Function(BaseRequest request);

/// Called when [response] is received, regardless of the statusCode. The returned response will be used
typedef OnResponse = FutureOr<StreamedResponse> Function(StreamedResponse response);

/// Called when statusCode represents a failure response. You can return a new request to retry or
/// null to throw the [error]
typedef OnError = FutureOr<BaseRequest?> Function(XHTTPException error);

/// Multipart request status callback. [fileIndex] is the index of the file being processed,
/// [length] is the current processed bytes and [total] is the overall length of the file
typedef OnMultipartFileProgress = void Function(int fileIndex, int length, int total);

/// Type alias for query parameters map
typedef QueryParameters = Map<String, dynamic>;

/// Type alias for headers map
typedef HeadersMap = Map<String, String>;

class XHttpClient with Loggable {
  final ValueGetter<Client> clientFactory;

  /// Used for building requests url
  final String? baseUrl;

  /// Creates a XHttpClient from [clientFactory], defaults to Client()
  XHttpClient({this.clientFactory = _defaultClientFactory, this.baseUrl});

  Future<Response> head(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('HEAD', buildUrl(url, params: params), headers: headers),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> get(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('GET', buildUrl(url, params: params), headers: headers),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> post(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Object? body,
    Encoding? encoding,
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('POST', buildUrl(url, params: params), headers: headers, body: body, encoding: encoding),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> put(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Object? body,
    Encoding? encoding,
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('PUT', buildUrl(url, params: params), headers: headers, body: body, encoding: encoding),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> patch(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Object? body,
    Encoding? encoding,
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('PATCH', buildUrl(url, params: params),
            headers: headers, body: body, encoding: encoding),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> delete(
    String url, {
    QueryParameters params = const {},
    HeadersMap headers = const {},
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) =>
      send(
        applicationJsonRequest('DELETE', buildUrl(url, params: params), headers: headers),
        onRequest: onRequest,
        onResponse: onResponse,
        onError: onError,
        cancel: cancel,
      );

  Future<Response> multipart(
    String url, {
    String method = 'POST',
    QueryParameters params = const {},
    HeadersMap headers = const {},
    Map<String, String> fields = const {},
    List<MultipartFile> files = const [],
    OnMultipartFileProgress? onProgress,
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) {
    BaseRequest request = MultipartRequest(method, buildUrl(url, params: params))
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'multipart/form-data')
      ..fields.addAll(fields)
      ..files.addAll(files);
    if (onProgress != null) {
      var byteCount = 0;
      var fileIndex = 0;
      request = request.observeProgress((readCount) {
        if (fileIndex >= files.length) return;
        byteCount += readCount;
        final length = files[fileIndex].length;
        if (byteCount > length) {
          onProgress(fileIndex, length, length);
          byteCount -= length;
          fileIndex++;
        } else {
          onProgress(fileIndex, byteCount, length);
        }
      });
    }
    return send(request, onRequest: onRequest, onResponse: onResponse, onError: onError, cancel: cancel);
  }

  Future<Response> send(
    BaseRequest request, {
    OnRequest? onRequest,
    OnResponse? onResponse,
    OnError? onError,
    CancellationToken? cancel,
  }) async {
    if (onRequest != null) request = await onRequest(request);
    _printRequest(request);
    final client = clientFactory();
    try {
      var stream = await CancellationToken.wrap(client.send(request), cancel);
      if (onResponse != null) stream = await onResponse(stream);
      final response = await Response.fromStream(stream);
      _printResponse(response);
      if (response.isSuccess) return response;
      final error = XHTTPException(response);
      if (onError != null) {
        final retry = await onError(error);
        if (retry != null) return send(retry);
      }
      throw error;
    } finally {
      client.close();
    }
  }

  Uri buildUrl(String url, {QueryParameters params = const {}}) =>
      buildUrlBase(baseUrl: baseUrl, url: url, params: params);

  void _printRequest(BaseRequest request) => logInfo('HTTP Request ${request.method} ${request.url}', () {
        _logKV('method', request.method);
        _logKV('url', '${request.url.scheme}://${request.url.host}:${request.url.port}${request.url.path}');
        final parameters = request.url.queryParameters;
        if (parameters.isNotEmpty) logDebug('parameters:', () => parameters.forEach(_logKV));
        if (request.headers.isNotEmpty) logDebug('headers:', () => request.headers.forEach(_logKV));
        if (request.contentLength != null) _logKV('contentLength', request.contentLength);
        void _printRequestBody(BaseRequest request) {
          if (request is Request) {
            final length = request.bodyBytes.length;
            if (length > 0)
              logDebug('body:', () {
                final type = request.contentType.mimeType;
                if (type == 'application/x-www-form-urlencoded')
                  request.bodyFields.forEach(_logKV);
                else if (const ['application/json', 'text/'].any(type.startsWith))
                  logDebug(request.body);
                else
                  logDebug('$length bytes');
              });
          } else if (request is StreamedRequest) {
            final length = request.contentLength;
            if (length == null || length > 0)
              logDebug('body:', () => logDebug(length == null ? 'stream' : '$length bytes stream'));
          } else if (request is MultipartRequest) {
            logDebug('body:', () {
              if (request.fields.isNotEmpty) logDebug('fields', () => request.fields.forEach(_logKV));
              if (request.files.isNotEmpty)
                logDebug('files', () => request.files.map((file) => file.filename ?? '').forEach(logDebug));
            });
          } else if (request is TransformRequest) {
            _logKV('transform', true);
            _printRequestBody(request.source);
          }
        }

        _printRequestBody(request);
      });

  void _printResponse(BaseResponse response) =>
      logInfo('HTTP Response ${response.request?.method} ${response.request?.url}', () {
        if (response.headers.isNotEmpty) logDebug('headers:', () => response.headers.forEach(_logKV));
        _logKV('statusCode', response.statusCode, level: LogLevel.info);
        if (response.reasonPhrase != null) _logKV('reasonPhrase', response.reasonPhrase, level: LogLevel.info);
        if (response is Response) {
          final length = response.bodyBytes.length;
          if (length > 0)
            logDebug('body:', () {
              final type = response.contentType.mimeType;
              if (const ['application/json', 'text/'].any(type.startsWith))
                logDebug(response.body);
              else
                logDebug('$length bytes');
            });
        } else if (response is StreamedResponse) {
          final length = response.contentLength;
          if (length == null || length > 0)
            logDebug('body:', () => logDebug(length == null ? 'stream' : '$length bytes stream'));
        }
      });

  void _logKV(dynamic key, dynamic value, {LogLevel level = LogLevel.debug}) => log(level, '$key: $value');

  /// Creates an HTTP application/json request
  ///
  /// The content-type of the request will default to 'application/json' if not set.
  ///
  /// If provided, [body] sets the body of the request. It can be a [String], any [Object].
  ///
  /// If [body] is a String, it's encoded using [encoding] and used as the body of the request
  /// else it's encoded to a json String through [jsonEncode] and treated as a String
  ///
  /// [encoding] defaults to [utf8].
  static Request applicationJsonRequest(String method, Uri uri,
      {HeadersMap headers = const {}, Object? body, Encoding? encoding}) {
    final request = Request(method, uri)
      ..headers.addAll(headers)
      ..headers.putIfAbsent('content-type', () => 'application/json');
    if (encoding != null) request.encoding = encoding;
    if (body != null) request.body = body is String ? body : jsonEncode(body);
    return request;
  }

  /// Build an uri from [baseUrl], [url] and query [params]
  ///
  /// Note: if [baseUrl] is not set or [url] does not start with '/', only [url] will be used.
  ///
  /// E.g. buildUrlBase(baseUrl: 'https://my.domain.com', url: '/api/myEndPoint', params: {'q': 'hello'})
  ///      -> 'https://my.domain.com/api/myEndPoint?q=hello'
  static Uri buildUrlBase({String? baseUrl, required String url, QueryParameters params = const {}}) {
    if (baseUrl != null && url.startsWith('/'))
      url = baseUrl.endsWith('/') ? baseUrl + url.substring(1) : baseUrl + url;
    final uri = Uri.parse(url);
    return params.isEmpty ? uri : uri.appendQueryParameters(params);
  }

  static Client _defaultClientFactory() => Client();
}
